--
-- Formspecs
--

function forestry_bees.get_apiary_active_formspec(queen_percent)
	return "size[8,8.5]"..
		"list[context;monarch;2.75,0.5;1,1;]"..
		"list[context;drone;2.75,2.5;1,1;]"..
		"image[3.75,1.5;1,1;gui_furnace_arrow_bg.png^[lowpart:"..
		(queen_percent)..":gui_furnace_arrow_fg.png^[transformR270]"..
		"list[context;dst;4.75,0.96;3,3;]"..
		"list[current_player;main;0,4.25;8,1;]"..
		"list[current_player;main;0,5.5;8,3;8]"..
		"listring[context;dst]"..
		"listring[current_player;main]"..
		"listring[context;src]"..
		"listring[current_player;main]"..
		"listring[context;fuel]"..
		"listring[current_player;main]"..
		default.get_hotbar_bg(0, 4.25)
end

function forestry_bees.get_apiary_inactive_formspec()
	return "size[8,8.5]"..
		"list[context;monarch;2.75,0.5;1,1;]"..
		"list[context;drone;2.75,2.5;1,1;]"..
		"image[3.75,1.5;1,1;gui_furnace_arrow_bg.png^[transformR270]"..
		"list[context;dst;4.75,0.96;3,3;]"..
		"list[current_player;main;0,4.25;8,1;]"..
		"list[current_player;main;0,5.5;8,3;8]"..
		"listring[context;dst]"..
		"listring[current_player;main]"..
		"listring[context;src]"..
		"listring[current_player;main]"..
		"listring[context;fuel]"..
		"listring[current_player;main]"..
		default.get_hotbar_bg(0, 4.25)
end

--
-- Node callback functions that are the same for active and inactive furnace
--

local function can_dig(pos, player)
	local meta = minetest.get_meta(pos);
	local inv = meta:get_inventory()
	return inv:is_empty("monarch") and inv:is_empty("drone") and inv:is_empty("dst")
end

local function allow_metadata_inventory_put(pos, listname, index, stack, player)
	if minetest.is_protected(pos, player:get_player_name()) then
		return 0
	end
	local meta = minetest.get_meta(pos)
	local inv = meta:get_inventory()
	if listname == "drone" then
    if minetest.get_item_group(stack:get_name(),"bee_drone") == 1 then
      return stack:get_count()
    else
			return 0
		end
	elseif listname == "monarch" then
    if minetest.get_item_group(stack:get_name(),"bee_monarch") == 1 then
      return stack:get_count()
    else
			return 0
		end
	elseif listname == "dst" then
		return 0
	end
end

local function allow_metadata_inventory_move(pos, from_list, from_index, to_list, to_index, count, player)
	local meta = minetest.get_meta(pos)
	local inv = meta:get_inventory()
	local stack = inv:get_stack(from_list, from_index)
	return allow_metadata_inventory_put(pos, to_list, to_index, stack, player)
end

local function allow_metadata_inventory_take(pos, listname, index, stack, player)
	if minetest.is_protected(pos, player:get_player_name()) then
		return 0
	end
	return stack:get_count()
end

local function swap_node(pos, name)
	local node = minetest.get_node(pos)
	if node.name == name then
		return
	end
	node.name = name
	minetest.swap_node(pos, node)
end

local function bee_change_type(bee_name,final_type)
	--given a bee_name it gives the same bee_name but with a different type, for example queen -> princess
	return string.gsub(bee_name,"_.-","_"..final_type)
end

local function apiary_result(queen)
	--given a queen calculates the correct princess + drone output
	--atm easy just returns the same princess and the same drone
	local meta = queen:get_meta()
	--the "meta" of the queen is the princess meta and the drone_"meta" is the drone meta
	local princess = Bee(bee_change_type(queen:get_name(),"princess"),meta:get_string("active_gene"),meta:get_string("inactive_gene"))
	local drone = Bee(bee_change_type(queen:get_name(),"drone"),meta:get_string("drone_active_gene"),meta:get_string("drone_inactive_gene"))
	drone:set_count(2)
	return princess, drone
end

local function apiary_node_timer(pos,elapsed)
    local meta = minetest.get_meta(pos)
    local src_time = meta:get_float("src_time") or 0  --how much time has passed since src started actioning
    local inv = meta:get_inventory()
    local monarchlist, dronelist
    local timer_elapsed = meta:get_int("timer_elapsed") or 0
    meta:set_int("timer_elapsed", timer_elapsed + 1)
    local update = true
    local is_queen, is_princess, is_drone
    local can_breed, can_live
    local actionable
    
    local breedtime = 5 --hardcoded for now
    
    while elapsed > 0 and update do
        update = false
        monarchlist = inv:get_list("monarch")
        dronelist = inv:get_list("drone")
        is_queen = minetest.get_item_group(monarchlist[1]:get_name(),"bee_queen") == 1
        is_princess = minetest.get_item_group(monarchlist[1]:get_name(),"bee_princess") == 1
        is_drone = minetest.get_item_group(dronelist[1]:get_name(),"bee_drone") == 1
        
        can_breed = is_princess and is_drone
        can_live = is_queen
        actionable = can_breed or can_live
        
		local el = elapsed
		if actionable then -- adjust el to action duration
			el = math.min(el, breedtime - src_time)
		end
		
		--if actionable check if things are ready
		if actionable then
			src_time = src_time + el
			if src_time >= breedtime then
				--if possible act accordingly to breed or live
				if can_breed then
					local queen = Bee(bee_change_type())
					inv:set_stack("monarch",1,)
				end
			end
		end
		
    end
end

local function furnace_node_timer(pos, elapsed)
	--
	-- Initialize metadata
	--
	local meta = minetest.get_meta(pos)
	--local fuel_time = meta:get_float("fuel_time") or 0
	local src_time = meta:get_float("src_time") or 0
	--local fuel_totaltime = meta:get_float("fuel_totaltime") or 0

	local inv = meta:get_inventory()
	local monarchlist, dronelist
	local dst_full = false

	local timer_elapsed = meta:get_int("timer_elapsed") or 0
	meta:set_int("timer_elapsed", timer_elapsed + 1)

	--local cookable, cooked
  local breedable
	--local fuel

	local update = true
	--local items_smelt = 0
	while elapsed > 0 and update do
		update = false

		monarchlist = inv:get_list("monarch")
		dronelist = inv:get_list("drone")

		--
		-- Cooking
		--

		-- Check if we have breedable content
		--local aftercooked
		--cooked, aftercooked = minetest.get_craft_result({method = "cooking", width = 1, items = srclist})
		breedable = (minetest.get_item_group(monarchlist[1]:get_name(),"bee_monarch") == 1) and (minetest.get_item_group(dronelist[1]:get_name(),"bee_drone") == 1)

		local el = math.min(elapsed, fuel_totaltime - fuel_time)
		if cookable then -- fuel lasts long enough, adjust el to cooking duration
			el = math.min(el, cooked.time - src_time)
		end

		-- Check if we have enough fuel to burn
		if fuel_time < fuel_totaltime then
			-- The furnace is currently active and has enough fuel
			fuel_time = fuel_time + el
			-- If there is a cookable item then check if it is ready yet
			if cookable then
				src_time = src_time + el
				if src_time >= cooked.time then
					-- Place result in dst list if possible
					if inv:room_for_item("dst", cooked.item) then
						inv:add_item("dst", cooked.item)
						inv:set_stack("src", 1, aftercooked.items[1])
						src_time = src_time - cooked.time
						update = true
					else
						dst_full = true
					end
					items_smelt = items_smelt + 1
				else
					-- Item could not be cooked: probably missing fuel
					update = true
				end
			end
		else
			-- Furnace ran out of fuel
			if cookable then
				-- We need to get new fuel
				local afterfuel
				fuel, afterfuel = minetest.get_craft_result({method = "fuel", width = 1, items = fuellist})

				if fuel.time == 0 then
					-- No valid fuel in fuel list
					fuel_totaltime = 0
					src_time = 0
				else
					-- prevent blocking of fuel inventory (for automatization mods)
					local is_fuel = minetest.get_craft_result({method = "fuel", width = 1, items = {afterfuel.items[1]:to_string()}})
					if is_fuel.time == 0 then
						table.insert(fuel.replacements, afterfuel.items[1])
						inv:set_stack("fuel", 1, "")
					else
						-- Take fuel from fuel list
						inv:set_stack("fuel", 1, afterfuel.items[1])
					end
					-- Put replacements in dst list or drop them on the furnace.
					local replacements = fuel.replacements
					if replacements[1] then
						local leftover = inv:add_item("dst", replacements[1])
						if not leftover:is_empty() then
							local above = vector.new(pos.x, pos.y + 1, pos.z)
							local drop_pos = minetest.find_node_near(above, 1, {"air"}) or above
							minetest.item_drop(replacements[1], nil, drop_pos)
						end
					end
					update = true
					fuel_totaltime = fuel.time + (fuel_totaltime - fuel_time)
				end
			else
				-- We don't need to get new fuel since there is no cookable item
				fuel_totaltime = 0
				src_time = 0
			end
			fuel_time = 0
		end

		elapsed = elapsed - el
	end

	if items_smelt > 0 then
		-- Play cooling sound
		minetest.sound_play("default_cool_lava",
			{ pos = pos, max_hear_distance = 16, gain = 0.07 * math.min(items_smelt, 7) }, true)
	end
	if fuel and fuel_totaltime > fuel.time then
		fuel_totaltime = fuel.time
	end
	if srclist and srclist[1]:is_empty() then
		src_time = 0
	end

	--
	-- Update formspec, infotext and node
	--
	local formspec
	local item_state
	local item_percent = 0
	if cookable then
		item_percent = math.floor(src_time / cooked.time * 100)
		if dst_full then
			item_state = S("100% (output full)")
		else
			item_state = S("@1%", item_percent)
		end
	else
		if srclist and not srclist[1]:is_empty() then
			item_state = S("Not cookable")
		else
			item_state = S("Empty")
		end
	end

	local fuel_state = S("Empty")
	local active = false
	local result = false

	if fuel_totaltime ~= 0 then
		active = true
		local fuel_percent = 100 - math.floor(fuel_time / fuel_totaltime * 100)
		fuel_state = S("@1%", fuel_percent)
		formspec = default.get_furnace_active_formspec(fuel_percent, item_percent)
		swap_node(pos, "default:furnace_active")
		-- make sure timer restarts automatically
		result = true

		-- Play sound every 5 seconds while the furnace is active
		if timer_elapsed == 0 or (timer_elapsed + 1) % 5 == 0 then
			local sound_id = minetest.sound_play("default_furnace_active",
				{pos = pos, max_hear_distance = 16, gain = 0.25})
			local hash = minetest.hash_node_position(pos)
			furnace_fire_sounds[hash] = furnace_fire_sounds[hash] or {}
			table.insert(furnace_fire_sounds[hash], sound_id)
			-- Only remember the 3 last sound handles
			if #furnace_fire_sounds[hash] > 3 then
				table.remove(furnace_fire_sounds[hash], 1)
			end
			-- Remove the sound ID automatically from table after 11 seconds
			minetest.after(11, function()
				if not furnace_fire_sounds[hash] then
					return
				end
				for f=#furnace_fire_sounds[hash], 1, -1 do
					if furnace_fire_sounds[hash][f] == sound_id then
						table.remove(furnace_fire_sounds[hash], f)
					end
				end
				if #furnace_fire_sounds[hash] == 0 then
					furnace_fire_sounds[hash] = nil
				end
			end)
		end
	else
		if fuellist and not fuellist[1]:is_empty() then
			fuel_state = S("@1%", 0)
		end
		formspec = default.get_furnace_inactive_formspec()
		swap_node(pos, "default:furnace")
		-- stop timer on the inactive furnace
		minetest.get_node_timer(pos):stop()
		meta:set_int("timer_elapsed", 0)

		stop_furnace_sound(pos)
	end


	local infotext
	if active then
		infotext = S("Furnace active")
	else
		infotext = S("Furnace inactive")
	end
	infotext = infotext .. "\n" .. S("(Item: @1; Fuel: @2)", item_state, fuel_state)

	--
	-- Set meta values
	--
	meta:set_float("fuel_totaltime", fuel_totaltime)
	meta:set_float("fuel_time", fuel_time)
	meta:set_float("src_time", src_time)
	meta:set_string("formspec", formspec)
	meta:set_string("infotext", infotext)

	return result
end

--
-- Node definitions
--

local function apply_logger(def)
	default.set_inventory_action_loggers(def, "apiary")
	return def
end

minetest.register_node("forestry_bees:apiary", apply_logger({
	description = S("Apiary"),
	tiles = {
		"forestry_bees_apiary_top.png", "forestry_bees_apiary_bottom.png",
		"forestry_bees_apiary_side.png", "forestry_bees_apiary_side.png",
		"forestry_bees_apiary_side.png", "forestry_bees_apiary_side.png"
	},
	paramtype2 = "facedir",
	groups = {choppy=2},
	legacy_facedir_simple = true,
	is_ground_content = false,
	sounds = default.node_sound_wood_defaults(),

	can_dig = can_dig,

	on_timer = apiary_node_timer,

	on_construct = function(pos)
		local meta = minetest.get_meta(pos)
		local inv = meta:get_inventory()
		inv:set_size('monarch', 1)
		inv:set_size('drone', 1)
		inv:set_size('dst', 9)
		apiary_node_timer(pos, 0)
	end,

	on_metadata_inventory_move = function(pos)
		minetest.get_node_timer(pos):start(1.0)
	end,
	on_metadata_inventory_put = function(pos)
		-- start timer function, it will sort out whether apiary can work or not.
		minetest.get_node_timer(pos):start(1.0)
	end,
	on_metadata_inventory_take = function(pos)
		-- check whether the apiary is empty or not.
		minetest.get_node_timer(pos):start(1.0)
	end,
	on_blast = function(pos)
		local drops = {}
		default.get_inventory_drops(pos, "monarch", drops)
		default.get_inventory_drops(pos, "drone", drops)
		default.get_inventory_drops(pos, "dst", drops)
		drops[#drops+1] = "forestry_bees:apiary"
		minetest.remove_node(pos)
		return drops
	end,

	allow_metadata_inventory_put = allow_metadata_inventory_put,
	allow_metadata_inventory_move = allow_metadata_inventory_move,
	allow_metadata_inventory_take = allow_metadata_inventory_take,
}))

--minetest.register_node("default:furnace_active", apply_logger({
	--description = S("Furnace"),
	--tiles = {
		--"default_furnace_top.png", "default_furnace_bottom.png",
		--"default_furnace_side.png", "default_furnace_side.png",
		--"default_furnace_side.png",
		--{
			--image = "default_furnace_front_active.png",
			--backface_culling = false,
			--animation = {
				--type = "vertical_frames",
				--aspect_w = 16,
				--aspect_h = 16,
				--length = 1.5
			--},
		--}
	--},
	--paramtype2 = "facedir",
	--light_source = 8,
	--drop = "default:furnace",
	--groups = {cracky=2, not_in_creative_inventory=1},
	--legacy_facedir_simple = true,
	--is_ground_content = false,
	--sounds = default.node_sound_stone_defaults(),
	--on_timer = furnace_node_timer,
	--on_destruct = function(pos)
		--stop_furnace_sound(pos)
	--end,

	--can_dig = can_dig,

	--allow_metadata_inventory_put = allow_metadata_inventory_put,
	--allow_metadata_inventory_move = allow_metadata_inventory_move,
	--allow_metadata_inventory_take = allow_metadata_inventory_take,
--}))

minetest.register_craft({
	output = "forestry_bees:apiary",
	recipe = {
		{"default:slab_wood", "default:slab_wood", "default:slab_wood"},
		{"default:wood", "", "default:wood"},
		{"default:wood", "default:wood", "default:wood"},
	}
})
